/*
 This is part of the OTF library. Copyright by ZIH, TU Dresden 2005-2009.
 Authors: Andreas Knuepfer, Denis Huenich, Johannes Spazier
*/

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif


#if defined(HAVE_SYS_TIME_H) && HAVE_SYS_TIME_H
#	include <sys/time.h>
#endif
#ifdef _WIN32
#	include <Winsock2.h>
#endif
#include <cassert>

#include "OTF_inttypes.h"
#include "OTF_Platform.h"

#include "Handler.h"
#include "Summary.h"
#include "Prodtex.h"
#include "CSVParse.h"


using namespace std;

#define HELPTEXT "" \
"                                                                  \n" \
" otfprofile -  generate a profile of a trace                      \n" \
"               in Latex or CSV format.                            \n" \
"                                                                  \n" \
"   options:                                                       \n" \
"      -h, --help    show this help message                        \n" \
"      -b <x>        readbuffer size                               \n" \
"      -f <x>        max. number of filehandles to use             \n" \
"      -i <file>     specify an input trace name                   \n" \
"      -csv <file>   specify an input csv-file trace name          \n" \
"                    (as produced by otfprofiler before),          \n" \
"                    don't use -i and -csv together                \n" \
"      -o <path>     specify the path for the output files         \n" \
"      -tex <x>      writes Latex output in different flavours:    \n" \
"                    (all,func,p2p,collop,none)                    \n" \
"      -notex        disable Latex output                          \n" \
"      -nops         disable Postscript output			   \n" \
"      -var          also show statistic variance                  \n" \
"      -top <x>      max. number of functions shown (default 50)   \n" \
"      -progress     show progress information                     \n" \
"      -sum          reads only summarized information, no events  \n" \
"                                                                  \n" \
 
int main( int argc, const char** argv ) 
{
	struct timeval tStart,tEnd;
	 	
	int i;
	OTF_FileManager* manager;
	OTF_Reader* reader;
	OTF_HandlerArray* handlers;
	int buffersize = 1024*1024;
	int nfiles = 10;
	string output_path("./");
	string output_dir;
	char* file = NULL;
	fstream FS;
	bool check = false; // if check == true then -i or -csv is set
	bool csv_on = false;
	vector<string> counter_names;
	global_data* gd = new global_data;
	bool sum = false;  // if true -> read summary

	/* show how many events were read */
	uint64_t read;
	uint64_t events = 0;
	uint64_t min;
	uint64_t cur;
	uint64_t max;
	double percent;
	
	/* what to read? */
	int  tex = TEX_ALL;
	bool status = false;
	bool file_rename = false;
	bool ps = true;
	gd->TOP_FUNC = 50;
	gd->var = false;
	gd->vis = false;
	gd->prog = false;
	gd->prog_start = 0;
	gd->prog_end = 0;
	gd->min_time = 0;
	gd->max_time = (uint64_t) -1;
	gd->ticks = 1;
	gd->clear_temp = false;

	if ( 1 >= argc ) {

		cout << HELPTEXT;
		return 1;
	}

	for(i = 1; i < argc; i++) 
	{
		if ((0 == strcmp("-b", argv[i])) && (i+1 < argc))
		{
			buffersize = atoi(argv[i+1]);
			++i;
		} 
		else if ((0 == strcmp("-tex", argv[i])) && (i+1 < argc))
		{
			if (0 == strcmp("all", argv[i+1]))
				tex = TEX_ALL;
			else if (0 == strcmp("allplot", argv[i+1]))
				tex = TEX_ALLPLOT;
			else if (0 == strcmp("func", argv[i+1]))
				tex = TEX_FUNC;
			else if (0 == strcmp("p2p", argv[i+1]))
				tex = TEX_P2P;
			else if (0 == strcmp("collop", argv[i+1]))
				tex = TEX_COLLOP;
			else if (0 == strcmp("none", argv[i+1]))
				tex = TEX_OFF;
			else
				cerr << "\nWrong argument given after -tex." << endl;
			++i;
		}
		else if ( 0 == strcmp("-notex", argv[i]) ) 
		{
			tex= TEX_OFF;
		}
		else if ((0 == strcmp("-f", argv[i])) && (i+1 < argc)) 
		{
			nfiles = atoi(argv[i+1]);
			++i;
		}
#if 0
		else if (0 == strcmp("-c", argv[i]) && (i+1 < argc)) 
		{
			i++;
			int test_end;
			while((0 != (test_end = strcmp("end", argv[i]))) && (i+1 < argc))
			{
				string s(strdup(argv[i]));
				transform(s.begin(), s.end(), s.begin(), ::toupper);
				counter_names.push_back(s);
				i++;
			}
			if(test_end != 0)
			{
				cerr << "Parameter -c : There was missing an \"end\"" << endl;
				exit(1);
			}
		}
#endif /* 0 */
		else if ((0 == strcmp("-min", argv[i])) && (i+1 < argc)) 
		{
			gd->min_time = atoi(argv[i+1]);
			++i;
		}
		else if ((0 == strcmp("-max", argv[i])) && (i+1 < argc)) 
		{
			gd->max_time = atoi(argv[i+1]);
			++i;
		}
		else if (0 == strcmp("-progress", argv[i])) 
		{
			status = true;
		}
		else if (0 == strcmp("--help", argv[i]) || 0 == strcmp("-h", argv[i])) 
		{
			cout << HELPTEXT;
			return 0;
		} 
		else if ((0 == strcmp("-i", argv[i])) && (i+1 < argc))
		{
			if(!check)
			{
				file = strdup(argv[i+1]);
				check = true;
			}
			++i;
		}
		else if ((0 == strcmp("-csv", argv[i])) && (i+1 < argc))
		{
			if(!check)
			{
				file = strdup(argv[i+1]);
				check = true;
				csv_on = true;
			}
			++i;
		}
		else if ((0 == strcmp("-o", argv[i])) && (i+1 < argc))
		{
			output_path = string(strdup(argv[i+1]));
			++i;
		} 
		else if ((0 == strcmp("-d", argv[i])) && (i+1 < argc))
		{
			output_path = string(strdup(argv[i+1]));
			file_rename = true;
			++i;
		}
		else if (0 == strcmp("-var", argv[i]))
		{
			gd->var = true;
		}
		else if (0 == strcmp("-vis", argv[i]))
		{
			gd->vis = true;
		}
		else if ((0 == strcmp("-top", argv[i])) && (i+1 < argc)) 
		{
			gd->TOP_FUNC = atoi(argv[i+1]);
			++i;
		}
		else if (0 == strcmp("-sum", argv[i]))
		{
			sum = true;
		}
		else if (0 == strcmp("-nops", argv[i]))
		{
			ps = false;
		}
		else 
		{
			cerr << "ERROR: Unknown argument: " <<  argv[i] << endl;;
			exit(1);
		}
	}
	
	if (nfiles < 1) 
	{
		cerr << "ERROR: need at least one input, aborting" << endl;
		exit(1);
	}
	
	if(file == NULL)
	{
		cerr << "ERROR: no file given" << endl;
		cerr << "To give a file, use \"-i filename\". " << endl;
		exit(1);
	}
	
	gd->filename.assign(OTF_basename(file));
	output_dir = output_path;
	if(!file_rename)
	{
		output_path.append(gd->filename);
	}
	gd->filename_path = output_path;
	
	if(csv_on)
	{
		Glob_Maps glob_maps;
		parse_csv(gd->sum_container, file, glob_maps);
		free(file);
	}
	else
	{	
		manager= OTF_FileManager_open(nfiles);
		assert(NULL != manager);

		handlers = OTF_HandlerArray_open();

		reader = OTF_Reader_open(file, manager);
		assert(NULL != reader);
	
		free(file);
	
		OTF_Reader_setBufferSizes(reader, buffersize);
	
		/* set the handler functions */

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefCreator, OTF_DEFCREATOR_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFCREATOR_RECORD);

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefVersion, OTF_DEFVERSION_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFVERSION_RECORD);

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefTimerResolution, OTF_DEFTIMERRESOLUTION_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFTIMERRESOLUTION_RECORD);
        
		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefFunction, OTF_DEFFUNCTION_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFFUNCTION_RECORD);

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefFunctionGroup, OTF_DEFFUNCTIONGROUP_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFFUNCTIONGROUP_RECORD);
		
		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefProcess, OTF_DEFPROCESS_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFPROCESS_RECORD);
		
		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefCollectiveOperation, OTF_DEFCOLLOP_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_DEFCOLLOP_RECORD);
        
   		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleDefCounter, OTF_DEFCOUNTER_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
      		(void*) gd, OTF_DEFCOUNTER_RECORD);
	

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleEnter, OTF_ENTER_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
      		(void*) gd, OTF_ENTER_RECORD);

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleLeave, OTF_LEAVE_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_LEAVE_RECORD);
        
   		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleCounter, OTF_COUNTER_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_COUNTER_RECORD);
		
		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleRecvMsg, OTF_RECEIVE_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
        	(void*) gd, OTF_RECEIVE_RECORD);

		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleSendMsg, OTF_SEND_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
      	  	(void*) gd, OTF_SEND_RECORD);
		
		OTF_HandlerArray_setHandler(handlers, 
			(OTF_FunctionPointer*) handleCollectiveOperation, OTF_COLLOP_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers, 
   	   	(void*) gd, OTF_COLLOP_RECORD);


		OTF_HandlerArray_setHandler(handlers,
			(OTF_FunctionPointer*) handleFunctionSummary, OTF_FUNCTIONSUMMARY_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers,
   	   	(void*) gd, OTF_FUNCTIONSUMMARY_RECORD);

		OTF_HandlerArray_setHandler(handlers,
			(OTF_FunctionPointer*) handleMessageSummary, OTF_MESSAGESUMMARY_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers,
   	   	(void*) gd, OTF_MESSAGESUMMARY_RECORD);

		OTF_HandlerArray_setHandler(handlers,
			(OTF_FunctionPointer*) handleCollopSummary, OTF_COLLOPSUMMARY_RECORD);
		OTF_HandlerArray_setFirstHandlerArg(handlers,
   	   	(void*) gd, OTF_COLLOPSUMMARY_RECORD);


		/* read statistics */

		if(sum) OTF_Reader_readStatistics(reader, handlers);

		/* read definitions */
		
		OTF_Reader_readDefinitions(reader, handlers);
		gd->sum_container.adddef_Bin(1);
		
		/* read events */
		
		if(!sum) {
		  if(status)
		  {
			gettimeofday(&tStart,0);
			cerr << "\n   starts reading the events" << endl;
			cerr.setf(ios::fixed, ios::floatfield);
			cerr.precision(2);
			OTF_Reader_setRecordLimit( reader, (uint64_t) 1000000 );
			cerr << endl;
		
			while(0 < (read = OTF_Reader_readEvents(reader, handlers))) 
			{
				events += read;
				OTF_Reader_eventProgress( reader, &min, &cur, &max );
				percent = 100.0 * ((double) (cur - min)) / ((double) (max - min));
				cerr <<  "   " << percent << "% of all events read, events read: " << events 
				     << "\r";
			}
		
			OTF_Reader_setRecordLimit( reader, (uint64_t) -1 );
			cerr << "\n   all events read\n" << endl;
			
			gettimeofday(&tEnd,0);
	
			cerr << "\n\n time used for profiling : " << 
		        (((double)(tEnd.tv_sec*1000000+tEnd.tv_usec) 
		        - (tStart.tv_sec*1000000+tStart.tv_usec))/1000000 ) << endl;
		  }
		  else  OTF_Reader_readEvents(reader, handlers);
		}
			
		set_time_sum_container(gd);
	
		FS.setf(ios::fixed, ios::floatfield);
		FS.precision(6);
		FS.open((gd->filename_path + "_func.csv").c_str(), ios::out | ios::trunc);
		gd->sum_container.csv_Function(FS, 1);
		FS.close();
		FS.open((gd->filename_path + "_p2p.csv").c_str(), ios::out | ios::trunc);
		gd->sum_container.csv_P2P(FS, 1);
		FS.close();
		FS.open((gd->filename_path + "_collop.csv").c_str(), ios::out | ios::trunc);
		gd->sum_container.csv_CollOp(FS, 1);
		FS.close();
		FS.open((gd->filename_path + "_data.csv").c_str(), ios::out | ios::trunc);
		gd->sum_container.csv_Data(FS, 1);
		FS.close();

		OTF_Reader_close(reader);
		OTF_HandlerArray_close(handlers);

		OTF_FileManager_close(manager);
	}

	if(tex != TEX_OFF) {
		prod_tex(tex, gd, counter_names, sum);
			
		if(ps == true) {
			string exe = "cd " + output_dir + "; latex " + gd->filename + "_result.tex > /dev/null 2>&1";
			int ret = system(exe.c_str());
			if(ret != 0) {
			  cout << " Warning: could not create ps-file! latex not installed." << endl;
			} else {
			  exe = "cd " + output_dir + "; dvips " + gd->filename + "_result.dvi > /dev/null 2>&1";
			  ret = system(exe.c_str());
			  if(ret != 0) {
			      cout << " Warning: could not create ps-file! dvips not installed." << endl;
			  }
			}
		}
	}

	return 0;
}
